import argparse
import logging
import sys
import os

from gen3schemadev.schema.gen3_template import (
    get_metaschema,
    generate_gen3_template,
    generate_def_template,
    generate_setting_template,
    generate_terms_template,
    generate_core_metadata_template,
    generate_project_template,
    generate_program_template
)
from gen3schemadev.utils import write_yaml, load_yaml, bundle_yamls, write_json, resolve_schema, read_json
from gen3schemadev.schema.input_schema import DataModel
from gen3schemadev.converter import get_node_names, populate_template
from gen3schemadev.validators.metaschema_validator import validate_schema_with_metaschema
from importlib.metadata import version
from gen3schemadev.ddvis import visualise_with_docker
from gen3schemadev.validators.rule_validator import RuleValidator


def main():
    version_parser = argparse.ArgumentParser(add_help=False)
    version_parser.add_argument(
        '--version',
        action='version',
        version=f"gen3schemadev {version('gen3schemadev')}"
    )

    version_parser.parse_known_args()

    parser = argparse.ArgumentParser(
        description="Gen3 Schema Development Tool"
    )
    parser.add_argument(
        '--version',
        action='version',
        version=f"%(prog)s {version('gen3schemadev')}"
    )

    subparsers = parser.add_subparsers(dest="command", required=False)

    # Create 'generate' subcommand
    generate_parser = subparsers.add_parser(
        "generate",
        help="Generate schemas"
    )
    generate_parser.add_argument(
        "-i", "--input",
        required=True,
        help="Input YAML file"
    )
    generate_parser.add_argument(
        "-o", "--output",
        required=True,
        help="Output directory"
    )
    generate_parser.add_argument(
        "--debug",
        action="store_true",
        help="Set logging level to DEBUG"
    )

    # Create 'bundle' subcommand
    bundle_parser = subparsers.add_parser(
        "bundle",
        help="Bundle YAML files"
    )
    bundle_parser.add_argument(
        "-i", "--input",
        required=True,
        help="Input directory"
    )
    bundle_parser.add_argument(
        "-f", "--filename",
        required=True,
        help="Output Filename"
    )
    bundle_parser.add_argument(
        "--debug",
        action="store_true",
        help="Set logging level to DEBUG"
    )

    # Create 'validate' subcommand
    validate_parser = subparsers.add_parser(
        "validate",
        help="Validate schemas"
    )
    validate_parser.add_argument(
        "-b", "--bundled",
        required=False,
        help="Bundled JsonSchema file"
    )
    validate_parser.add_argument(
        "-y", "--yamls",
        required=False,
        help="Directory of Gen3 Yaml files"
    )
    validate_parser.add_argument(
        "--debug",
        action="store_true",
        help="Set logging level to DEBUG"
    )
    validate_parser.add_argument(
        "--no-exclude",
        action="store_true",
        help="Disables the exclusion of specific schema from the validation"
    )

    # Create 'visualise' subcomand
    visualise_parser = subparsers.add_parser(
        "visualise",
        help="Visualise schemas"
    )
    visualise_parser.add_argument(
        "-i", "--input",
        required=True,
        help="Path to Bundled Gen3 JsonSchema file"
    )
    visualise_parser.add_argument(
        "--debug",
        action="store_true",
        help="Set logging level to DEBUG"
    )

    # Create 'init' subcommand
    init_parser = subparsers.add_parser(
        "init",
        help="Initialize gen3schemadev"
    )
    init_parser.add_argument(
        "-o", "--output",
        required=False,
        help="Output directory"
    )
    init_parser.add_argument(
        "--debug",
        action="store_true",
        help="Set logging level to DEBUG"
    )
    
    args = parser.parse_args()
    
    # Handle case where no command is provided
    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(0)

    # Set up basic logging configuration
    # If the subcommand has --debug, set to DEBUG, else INFO
    log_level = logging.ERROR
    if hasattr(args, "debug") and getattr(args, "debug", False):
        log_level = logging.DEBUG
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s [%(levelname)s] %(message)s"
    )
    logger = logging.getLogger(__name__)

    if args.command == "generate":
        print("Starting schema generation process...")
        metaschema = get_metaschema()
        converter_template = generate_gen3_template(metaschema)
        print(f"Loading input YAML from: {args.input}")
        data = load_yaml(args.input)
        print("Validating input data model...")
        validated_model = DataModel.model_validate(data)
        node_names = get_node_names(validated_model)
        print(f"Found nodes: {node_names}")

        for node in node_names:
            print(f"Populating template for node: '{node}'")
            out_template = populate_template(node, validated_model, converter_template)
            print(f"Writing output YAML to: {args.output}/{node}.yaml")
            write_yaml(out_template, f"{args.output}/{node}.yaml")

        print('Writing auxiliary files')
        write_yaml(generate_def_template(), f"{args.output}/_definitions.yaml")
        setting_dict = generate_setting_template()
        setting_dict['_dict_version'] = validated_model.version
        write_yaml(setting_dict, f"{args.output}/_settings.yaml")
        write_yaml(generate_terms_template(), f"{args.output}/_terms.yaml")
        write_yaml(generate_core_metadata_template(), f"{args.output}/core_metadata_collection.yaml")
        if 'project' not in node_names:
            write_yaml(generate_project_template(), f"{args.output}/project.yaml")
        if 'program' not in node_names:
            write_yaml(generate_program_template(), f"{args.output}/program.yaml")

        print("Schema generation process complete.")

    elif args.command == "bundle":
        print(f"Bundling YAML files from directory: {args.input}")
        bundle_dict = bundle_yamls(args.input)
        print(f"Writing bundled schema to file: {args.filename}")
        write_json(bundle_dict, args.filename)
        print("Bundling process complete.")

    elif args.command == "validate":
        print("Starting validation process...")
        metaschema = get_metaschema()

        # Exclusion list
        exclude_schema_list = [
            'project', '_definitions', '_settings',
            '_terms', 'core_metadata_collection',
            'program'
        ]
        if args.no_exclude:
            print(f"Validation now includes: {exclude_schema_list}")
            exclude_schema_list = []

        # Conducting business rule validation
        if args.bundled:
            schema_dict = read_json(args.bundled)
        elif args.yamls:
            schema_dict = bundle_yamls(args.yamls)

        for schema_name, schema in schema_dict.items():

            if '.' in schema_name:
                schema_name = os.path.splitext(schema_name)[0]

            if schema_name in exclude_schema_list:
                continue
  
            rule_validator = RuleValidator(schema)
            rule_validator.validate()
            print(f"SUCCESS: Rule validation complete for: {schema_name}")

        # Resolving bundled schema which is required for metaschema validation
        resolved_schema_dict = None

        if args.bundled:
            print(f"Resolving schema from bundled file: {args.bundled}")
            resolved_schema_dict = resolve_schema(schema_path=args.bundled)
        elif args.yamls:
            print(f"Bundling and resolving schemas from directory: {args.yamls}")
            resolved_schema_dict = resolve_schema(schema_dir=args.yamls)

        if resolved_schema_dict is None:
            logger.error("You must provide either --bundled or --yamls for validation.")
            sys.exit(1)


        for schema_name, schema in resolved_schema_dict.items():
            validate_schema_with_metaschema(
                schema,
                metaschema=metaschema,
                verbose=True
            )
            print(f"SUCCESS: Metaschema validation complete for: {schema_name}")

        print("Validation process complete.")

    elif args.command == "visualise":
        print(f"Visualising schema from file: {args.input}")
        visualise_with_docker(args.input)
    
    
    elif args.command == "init":
        init_yaml = {
            "version": "0.1.0",
            "url": "https://link-to-data-portal",
            "nodes": [
                {
                    "name": "subject",
                    "description": "The subject or patient",
                    "category": "clinical",
                    "properties": [
                        {
                            "name": "subject_id",
                            "type": "string",
                            "description": "A deidentified identifier for the subject",
                            "required": True
                        }
                    ]
                },
                {
                    "name": "sample",
                    "description": "Info about sample",
                    "category": "clinical",
                    "properties": [
                        {
                            "name": "sample_id",
                            "type": "string",
                            "description": "Sample ID (string)",
                            "required": True
                        }
                    ]
                }
            ],
            "links": [
                {
                    "parent": "subject",
                    "multiplicity": "one_to_many",
                    "child": "sample"
                }
            ]
        }

        output_path = args.output or "input_example.yaml"
        write_yaml(init_yaml, output_path)

if __name__ == "__main__":
    main()



# Generates gen3 jsonschema template for a single node using the gen3 metaschema
from gen3schemadev.utils import *
import os
import yaml
import importlib.resources
import logging

logger = logging.getLogger(__name__)

def read_template_yaml(template_filename='template.yml'):
    """
    Reads a YAML template file from the schema_templates directory.

    Args:
        template_filename (str): The name of the YAML template file.

    Returns:
        dict: The loaded YAML data as a dictionary.
    """
    current_dir = os.path.dirname(__file__)
    template_path = os.path.join(current_dir, 'schema_templates', template_filename)
    with open(template_path, 'r') as file:
        data = yaml.safe_load(file)
    return data

def generate_gen3_template(metaschema: dict) -> dict:
    """
    Generates a Gen3 JSON schema template for a single node using the provided metaschema YAML file.

    Args:
        metaschema (str): Path to the metaschema YAML file.

    Returns:
        dict: A dictionary containing the default values for each property defined in the metaschema.

    Raises:
        FileNotFoundError: If the metaschema file does not exist.
        yaml.YAMLError: If there is an error parsing the YAML file.
        Exception: For any other unexpected errors.
    """
    try:
        metaschema_data = metaschema
        out_template = {}
        properties = metaschema_data.get('properties', {})
        out_template['$schema'] = metaschema_data.get('$schema')
        logger.info(f"Generating Gen3 template from metaschema with {len(properties)} properties.")
        for k, v in properties.items():
            if 'default' in v:
                out_template[k] = v['default']
                logger.debug(f"Set default for property '{k}': {v['default']}")
            else:
                out_template[k] = None
                logger.debug(f"No default for property '{k}', set to None.")
        logger.info("Gen3 template generation completed successfully.")
        return out_template
    except Exception as e:
        logger.error(f"An error occurred while generating the Gen3 template: {e}")
        raise


def generate_def_template():
    return read_template_yaml('_definitions.yaml')

def generate_setting_template():
    return read_template_yaml('_settings.yaml')

def generate_terms_template():
    return read_template_yaml('_terms.yaml')

def generate_project_template():
    return read_template_yaml('project.yaml')

def generate_core_metadata_template():
    return read_template_yaml('core_metadata_collection.yaml')

def get_metaschema():
    return read_template_yaml('gen3_metaschema.yml')

def generate_program_template():
    return read_template_yaml('program.yaml')


# Expose main submodules for convenient import
from .schema import *
from .validators import *
from .validators.metaschema_validator import *
from .validators.rule_validator import *
from .validators.input_validator import *
from .cli import *
from .utils import *
from .converter import *
from .ddvis import *

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError  # for Python<3.8

try:
    from importlib.metadata import version
    __version__ = version('gen3schemadev')
except Exception:
    __version__ = 'unknown'


# validates gen3 bundled jsonschema (.json) by testing gen3 specific business rules
import logging
from gen3schemadev.converter import link_suffix

logger = logging.getLogger(__name__)


class RuleValidator:
    def __init__(self, schema: dict):
        self.schema = schema

    def validate(self):
        self.data_file_link_core_metadata()
        self.link_props_exist()
        self.props_cannot_be_system_props()
        self.props_must_have_type()
        self.type_array_needs_items()
        self.core_metadata_required_link()
        self.data_file_props_need_data_props()

    def _get_links(self):
        links = self.schema.get("links", [])
        logger.debug(f"Fetched links: {links}")
        if links and 'subgroup' in links[0]:
            logger.debug(
                f"Links contains 'subgroup', returning subgroup: {links[0]['subgroup']}"
            )
            return links[0]['subgroup']
        return links

    def _get_props(self):
        props = self.schema.get("properties", [])
        logger.debug(f"Fetched properties: {props}")
        return props

    def data_file_link_core_metadata(self):
        """
        If the schema has category: data_file,
        it must have a core metadata link.
        Raises informative exceptions when rule is violated.
        """
        try:
            category = self.schema.get("category")
            logger.debug(
                f"Checking if schema category is 'data_file'. Found: {category}"
            )
            schema_id = self.schema.get("id", "<unknown id>")
            if category == "data_file":
                links = self._get_links()
                logger.debug(f"Got links for data_file: {links}")
                for link in links:
                    if link.get("name") == link_suffix("core_metadata_collection"):
                        logger.debug(
                            f"Found core_metadata_collection link for data_file node (id: {schema_id})."
                        )
                        return True
                logger.warning(
                    f"No core_metadata_collection link found for data_file node (id: {schema_id})."
                )
                raise ValueError(
                    f"Schema '{schema_id}' with category 'data_file' must include a link with "
                    f"'name': 'core_metadata_collections'. Please add this link to the 'links' section."
                )
            return False
        except Exception as ex:
            schema_id = self.schema.get("id", "<unknown id>")
            raise RuntimeError(
                f"Exception occurred while validating core_metadata_collection link for schema '{schema_id}': {ex}"
            ) from ex

    def link_props_exist(self):
        """
        Ensures that properties exist for each link defined in the schema.

        Raises:
            ValueError: If a property corresponding to a link is missing, includes schema id
                        and offending link name for user troubleshooting.
            Exception: For unexpected errors.
        """
        try:
            links = self._get_links()
            link_name_list = []
            schema_id = self.schema.get("id", "<unknown id>")
            for link in links:
                link_name_list.append(link.get("name"))
            logger.debug(f"Collected link names for schema '{schema_id}': {link_name_list}")

            props = self._get_props()
            prop_keys = list(props.keys())
            logger.debug(f"Collected property keys for schema '{schema_id}': {prop_keys}")
            for link_name in link_name_list:
                if link_name not in prop_keys:
                    logger.error(
                        f"Property for link '{link_name}' does not exist in properties of schema '{schema_id}'."
                    )
                    raise ValueError(
                        f"In schema '{schema_id}', property for link '{link_name}' "
                        f"is missing from the 'properties' section. "
                        f"Please add a property named '{link_name}' to resolve this."
                    )
            logger.debug(
                f"All link names have corresponding properties for schema '{schema_id}'."
            )
            return True
        except Exception as ex:
            schema_id = self.schema.get("id", "<unknown id>")
            raise RuntimeError(
                f"Exception while verifying link properties for schema '{schema_id}': {ex}"
            ) from ex

    def _system_props(self):
        """returns a list of system properties 
        that you gen3 uses on the backend. You
        should not name any of your properties these.
        """
        system_props = [
            "id",
            "project_id",
            "type",
            "submitter_id",
            "state",
            "created_datetime",
            "updated_datetime",
            "file_state",
            "error_type",
            "label"
        ]
        logger.debug(f"System properties: {system_props}")
        return system_props

    def props_cannot_be_system_props(self):
        props = self._get_props()
        prop_keys = list(props.keys())
        schema_id = self.schema.get("id", "<unknown id>")
        logger.debug(
            f"Checking user property keys against system properties for schema '{schema_id}': {prop_keys}"
        )
        try:
            for prop in prop_keys:
                if prop in self._system_props():
                    logger.error(
                        f"Property '{prop}' should not use a reserved system name in schema '{schema_id}'."
                    )
                    raise ValueError(
                        f"In schema '{schema_id}', property '{prop}' uses a reserved system property name. "
                        f"Please rename this property to avoid conflicts with internal Gen3 system fields."
                    )
            logger.debug(
                f"No property keys found that overlap with system properties in schema '{schema_id}'."
            )
            return True
        except Exception as ex:
            raise RuntimeError(
                f"Exception while checking for system property key conflicts in schema '{schema_id}': {ex}"
            ) from ex


    def props_must_have_type(self):
        """
        Ensure all user-defined properties (that are dictionaries and not only $ref)
        have a "type" or "enum" key.
        """
        props = self._get_props()
        schema_id = self.schema.get("id", "<unknown id>")

        for key, value in props.items():
            # Skip schema-level $ref
            if key == "$ref":
                logger.debug(
                    f"Skipping property '{key}' because it is a schema reference."
                )
                continue

            # Require property definition to be a dict
            if not isinstance(value, dict):
                logger.debug(
                    f"Skipping property '{key}' because it is not a dictionary."
                )
                continue

            # Skip if property definition contains a $ref (i.e., is an alias/reference)
            if "$ref" in value:
                logger.debug(
                    f"Skipping property '{key}' because it contains a '$ref'."
                )
                continue

            # Require at least 'type' or 'enum' to be present
            if "type" not in value and "enum" not in value:
                logger.error(
                    f"Property '{key}' in schema '{schema_id}' is missing a 'type' or 'enum' field."
                )
                raise ValueError(
                    f"Property '{key}' must have a value for 'type' or 'enum' in schema '{schema_id}'."
                )
        return True
    
    def data_file_props_need_data_props(self):
        """If the schema is category: data_file, then it must 
        have the properties `data_type`, `data_format`, and `data_category`.
        """

        if not self.schema.get('category', '') == 'data_file':
            return True

        props = self._get_props()
        prop_keys = set(props.keys())
        need_props = {"data_type", "data_format", "data_category"}
        if not need_props.issubset(prop_keys):
            schema_id = self.schema.get("id", "<unknown id>")
            raise ValueError(
                f"Schema '{schema_id}' with category 'data_file' must include properties "
                f"'data_type', 'data_format', and 'data_category'. Please add these properties "
                f"to the 'properties' section."
            )
        return True

    
    def type_array_needs_items(self):
        """Ensure that any property with 'type': 'array' includes an 'items' property."""
        props = self._get_props()

        for key, value in props.items():
            if isinstance(value, dict) and value.get("type") == "array":
                if "items" not in value:
                    raise ValueError(
                        f"Schema '{self.schema.get('id')}' property '{key}' with type 'array' must include an 'items' property. "
                        "Please add an 'items' property to the 'properties' section."
                    )
        return True


    def core_metadata_required_link(self):
        """Core metadata collection schema needs at least one required link
        """
        if not self.schema.get('id', '') == 'core_metadata_collection':
            return True

        links = self._get_links()
        for link in links:
            val = link.get("required", None)
            if val is not None and val is True:
                return True
        # If loop completes without returning, then no required link exists
        raise ValueError(
            f"Schema '{self.schema.get('id')}' must include at least one required link. "
            "Please add a required link to the 'links' section."
        )



"""
Gen3 Schema Generation Module

This module provides utilities for converting node and link data into Gen3 schema format.
It handles node properties, links, and special cases like file nodes that require
core metadata collections.

Expected Data Structure:
    The input data should be a Pydantic model with:
    - nodes: List of node objects with attributes: name, description, category, properties, links
    - links: List of link objects with attributes: child, parent, multiplicity
"""

from dataclasses import dataclass, asdict
from typing import Protocol, runtime_checkable, Dict, Any, List
import logging


logger = logging.getLogger(__name__)


# Protocol definitions for type safety
@runtime_checkable
class nodeProtocol(Protocol):
    """Protocol defining the structure of an node object."""
    name: str
    description: str
    category: str
    properties: list
    links: list

    def model_dump(self) -> dict:
        """Convert node to dictionary."""
        ...


@runtime_checkable
class LinkProtocol(Protocol):
    """Protocol defining the structure of a link object."""
    child: str
    parent: str
    multiplicity: str

    def model_dump(self) -> dict:
        """Convert link to dictionary."""
        ...


@runtime_checkable
class DataSourceProtocol(Protocol):
    """Protocol defining the expected structure of input data."""
    nodes: list[nodeProtocol]
    links: list[LinkProtocol]


# Data classes
@dataclass
class node:
    """Represents a Gen3 node with its metadata and relationships."""
    name: str
    description: str
    category: str
    properties: list
    links: list

    def to_dict(self) -> dict:
        """Convert node to dictionary representation."""
        return asdict(self)


@dataclass
class LinkObj:
    """Represents a link between two nodes in Gen3 schema."""
    name: str
    backref: str
    label: str | None
    target_type: str
    multiplicity: str
    required: bool

    def to_dict(self) -> dict:
        """Convert link to dictionary representation."""
        return asdict(self)


@dataclass
class LinkGroup:
    """Represents a group of links with specific constraints."""
    exclusive: bool
    required: bool
    subgroup: list[dict]

    def to_dict(self) -> dict:
        """Convert link group to dictionary representation."""
        return asdict(self)


# Utility functions
def link_suffix(word: str, suffix='s') -> str:
    """
    Adds link suffix to a singular word.

    Args:
        word: The singular word to suffix

    Returns:
        The work with the suffix
    """
    return word + suffix

def get_node_names(data: DataSourceProtocol) -> list[str]:
    """
    Retrieve a list of node names from the data structure.

    Args:
        data: The data structure containing nodes.

    Returns:
        A list of node names.
    """
    return [node.name for node in data.nodes]


def get_node_data(node: str, data: DataSourceProtocol) -> nodeProtocol:
    """
    Retrieve the node object for a given node name from the data structure.

    Args:
        node: The name of the node to retrieve.
        data: The data structure containing nodes.

    Returns:
        The node object matching the given name.

    Raises:
        ValueError: If the node is not found in data.nodes.
        AttributeError: If the data structure is invalid.
    """
    if not hasattr(data, 'nodes'):
        raise AttributeError("Data structure missing 'nodes' attribute")

    try:
        for ent in data.nodes:
            if ent.name == node:
                return ent
        raise ValueError(f"node '{node}' not found in data.nodes")
    except AttributeError as e:
        raise AttributeError(f"Invalid data structure: {e}")


def get_node_links(node: str, data: DataSourceProtocol) -> list[dict]:
    """
    Retrieve all links where the given node is the child.

    Args:
        node: The name of the node (child) to find links for.
        data: The data structure containing links.

    Returns:
        A list of link dictionaries where the node is the child.

    Raises:
        AttributeError: If the data structure is invalid.
    """
    if not hasattr(data, 'links'):
        raise AttributeError("Data structure missing 'links' attribute")

    links = data.links
    node_links = []
    for link in links:
        if link.child == node:
            node_links.append(link.model_dump())
    return node_links


def create_core_metadata_link(child_name: str, required: bool = False) -> dict:
    """
    Create a link dictionary for core metadata collections.

    This is used for file nodes that must be linked to a core_metadata_collection.

    Args:
        child_name: The name of the child node.

    Returns:
        A dictionary representing the core metadata link.
    """
    link_obj = LinkObj(
        name=link_suffix("core_metadata_collection"),
        backref=link_suffix(child_name),
        label="part_of",
        target_type="core_metadata_collection",
        multiplicity="one_to_one",
        required=required
    )
    return link_obj.to_dict()


def flip_multiplicity(multiplicity: str) -> str:
    """
    Flip the multiplicity of a link.

    Args:
        multiplicity: The current multiplicity of the link.

    Returns:
        The flipped multiplicity.
    """
    if multiplicity == "one_to_one" or multiplicity == "many_to_many":
        return multiplicity
    elif multiplicity == "one_to_many":
        return "many_to_one"
    elif multiplicity == "many_to_one":
        return "one_to_many"
    else:
        raise ValueError(f"Invalid multiplicity: {multiplicity}")

def convert_node_links(links: list[dict], required: bool = True) -> list[dict]:
    """
    Convert a list of link dictionaries into the Gen3 schema 'links' format.

    Args:
        links: A list of link dictionaries for the node.
        required: Whether the links are required (default: True).

    Returns:
        A list of formatted link dictionaries for the Gen3 schema.
    """
    link_list = []
    for link in links:
        link_obj = LinkObj(
            name=link_suffix(link['parent']),
            backref=link_suffix(link['child']),
            label="part_of",
            target_type=link['parent'],
            multiplicity=flip_multiplicity(link['multiplicity']),
            required=required
        )
        link_list.append(link_obj.to_dict())

    return link_list


def add_core_metadata_link(links: list[dict], child_name: str) -> list[dict]:
    """
    Add a core metadata link to an existing list of links for file nodes.

    Args:
        links: Existing list of link dictionaries.
        child_name: The name of the child node.

    Returns:
        Updated list of link dictionaries including core metadata link.
    """
    core_link = create_core_metadata_link(child_name, required=False)
    return links + [core_link]


def create_link_group(links: list[dict], exclusive: bool = False, required: bool = True) -> list[dict]:
    """
    Create a link group structure from a list of links.

    Args:
        links: List of link dictionaries to group.
        exclusive: Whether the links are mutually exclusive.
        required: Whether at least one link is required.

    Returns:
        A list containing a dictionary representing the link group.
    """
    group = LinkGroup(
        exclusive=exclusive,
        required=required,
        subgroup=links
    )
    return [group.to_dict()]


def format_multiplicity(multiplicity: str) -> str:
    """
    For gen3 link properties, definitions for one_to_many links 
    are referenced as to_many. The def, says it is an array with
    a min length of 1. And for one_to_one, it is referenced as to_one
    with a max array length of 1. It does not matter if it is many_to 
    or one_to since that is implied by the number of submitter_ids for 
    the node.

    Args:
        multiplicity: The multiplicity of the link (e.g., 'one_to_one', 'one_to_many').

    Returns:
        The formatted multiplicity string.
    """
    # Map all *_to_one to "to_one", all *_to_many to "to_many"
    if not isinstance(multiplicity, str):
        logger.error(f"Multiplicity must be a string, got {type(multiplicity)}")
        raise ValueError(f"Multiplicity must be a string, got {type(multiplicity)}")
    if multiplicity.endswith("_to_one"):
        return "to_one"
    elif multiplicity.endswith("_to_many"):
        return "to_many"
    else:
        logger.error(f"Invalid multiplicity: {multiplicity}")
        raise ValueError(f"Invalid multiplicity: {multiplicity}")

def create_link_prop(target_node: str, multiplicity: str) -> dict:
    """
    Create a property dictionary for a link to another node.

    Args:
        target_node: The name of the target node.
        multiplicity: The multiplicity of the link (e.g., 'one_to_one', 'one_to_many').

    Returns:
        A dictionary representing the link property for the schema.
    """
    link_prop = {
        link_suffix(target_node): {
            "$ref": f"_definitions.yaml#/{format_multiplicity(multiplicity)}"
        }
    }
    return link_prop


def get_properties(node_name: str, data: DataSourceProtocol) -> list[dict]:
    """
    Retrieve the list of property dictionaries for a given node.

    Args:
        node_name: The name of the node.
        data: The data structure containing nodes.

    Returns:
        A list of property dictionaries, or an empty list if no properties found.

    Raises:
        ValueError: If the node is not found.
    """
    ent = get_node_data(node_name, data)
    props = ent.properties
    
    output = []
    if props:
        for prop in props:
            pdict = {
                prop.name: {k: v for k, v in prop.model_dump().items() if k != "name"}
            }
            
            output.append(pdict)
            
    else:
        logger.debug(f"No properties found for node '{node_name}'")
    
    return output


def strip_required_field(props_list: list[dict]) -> list[dict]:
    """
    Remove the 'required' field from all property dicts in the input list.
    Can use the output of get_properties() for this function

    Args:
        props_list (list): A list of property dictionaries, where each dictionary has a single key
            (the property name) and its value is a dictionary describing the property. For example:
                [
                    {
                        "project_id": {
                            "type": "string",
                            "description": "Synthetic_Dataset_1",
                            "required": True,
                            "enums": None
                        }
                    },
                    ...
                ]

    Returns:
        list: A new list with the same structure as props_list, but with the 'required'
            field removed from each property's dictionary (if present).

    Note:
        This function expects a list of property definitions as typically returned by
        get_properties() in the Gen3 schema conversion workflow.
        If you are working with a DataSourceProtocol object, you should first extract the
        properties list using the appropriate function.
    """
    new_list = []
    for prop in props_list:
        if isinstance(prop, dict):
            # Each prop is {property_name: property_dict}
            new_prop = {}
            for k, v in prop.items():
                if isinstance(v, dict):
                    v = {key: val for key, val in v.items() if key != 'required'}
                new_prop[k] = v
            new_list.append(new_prop)
        else:
            new_list.append(prop)
    return new_list

def get_required_prop_names(props_list: list[dict]) -> List[str]:
    """
    Given a list of property dicts (as from get_properties), return a list of property names
    where the property dict has 'required': True. Can use the output of get_properties() for this function.

    Args:
        props_list (list): A list of property dictionaries, where each dictionary has a single key
            (the property name) and its value is a dictionary describing the property. For example:
                [
                    {
                        "project_id": {
                            "type": "string",
                            "description": "Synthetic_Dataset_1",
                            "required": True,
                            "enums": None
                        }
                    },
                    ...
                ]

    Returns:
        List[str]: List of property names with required True.
    """
    required_names = []
    for prop in props_list:
        if isinstance(prop, dict):
            for k, v in prop.items():
                if isinstance(v, dict) and v.get("required") is True:
                    required_names.append(k)
    return required_names


def format_enum(prop_dict: dict) -> dict:
    """
    Format an enum property dictionary for use in a Gen3 schema.

    Example:
        {'sample_tube_type': {'type': 'enum',
           'description': 'Sample tube type (enum)',
           'required': False,
           'enums': ['EDTA', 'Heparin', 'Citrate']
           }
        }

    Args:
        prop_dict (dict): The dictionary representing the enum property.

    Returns:
        dict: The formatted dictionary for use in a Gen3 schema.
    """
    try:
        if len(prop_dict) != 1:
            raise ValueError("Expected a single property dictionary")
        
        first_key = next(iter(prop_dict))
        value = prop_dict[first_key]
        formatted_props = {}
        
        if 'enums' in value and value['enums'] is not None:
            formatted_props['description'] = value['description']
            formatted_props['enum'] = value['enums']
        else:
            # Remove the 'enums' key from value if present
            value = value.copy()
            value.pop('enums', None)
            formatted_props = value
        
        output = {first_key: formatted_props}
        return output
    except Exception as e:
        # You may want to log the error or handle it differently depending on your use case
        raise RuntimeError(f"Error formatting enum property: {e}") from e


def format_datetime(prop_dict: dict) -> dict:
    """
    Formats a property dictionary with a 'type' of 'datetime' to use a $ref
    to the Gen3 _definitions.yaml#/datetime definition.

    If the property is not of type 'datetime', returns the property unchanged.

    Example input:
        {
            "collection_date": {
                "type": "datetime",
                "description": "Date and time of collection (datetime)"
            }
        }

    Example output:
        {
            "collection_date": {
                "$ref": "_definitions.yaml#/datetime"
            }
        }

    Args:
        prop_dict (dict): A dictionary with a single property as key and its attributes as value.

    Returns:
        dict: The formatted property dictionary suitable for Gen3 schema usage.

    Raises:
        ValueError: If the input dictionary does not contain exactly one property.
        RuntimeError: If an error occurs during formatting.
    """
    try:
        if len(prop_dict) != 1:
            raise ValueError("Expected a single property dictionary")
        
        first_key = next(iter(prop_dict))
        value = prop_dict[first_key]
        formatted_props = {}
        
        if 'type' in value and value['type'] == 'datetime':
            formatted_props['$ref'] = "_definitions.yaml#/datetime"
        else:
            formatted_props = value
        
        output = {first_key: formatted_props}
        return output
    except Exception as e:
        # You may want to log the error or handle it differently depending on your use case
        raise RuntimeError(f"Error formatting datetime property: {e}") from e


def format_array(prop_dict: dict) -> dict:
    """
    Formats a property dictionary with a 'type' of 'array' to use a Gen3 'array' schema definition.

    If the property is not of type 'array', returns the property unchanged.

    Example input:
        {
            "sample_ids": {
                "type": "array",
                "description": "Sample IDs (array)"
            }
        }

    Example output:
        {
            "sample_ids": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            }
        }

    Args:
        prop_dict (dict): A dictionary with a single property as key and its attributes as value.

    Returns:
        dict: The formatted property dictionary suitable for Gen3 schema usage.

    Raises:
        ValueError: If the input dictionary does not contain exactly one property.
        RuntimeError: If an error occurs during formatting.
    """
    try:
        if len(prop_dict) != 1:
            raise ValueError("Expected a single property dictionary")

        key, value = next(iter(prop_dict.items()))

        if isinstance(value, dict) and value.get('type') == 'array':
            # Safely preserve description and required if present
            formatted = {
                "type": "array",
                "items": {
                    "type": "string"
                }
            }
            if "description" in value:
                formatted["description"] = value["description"]
            if "required" in value:
                formatted["required"] = value["required"]
            return {key: formatted}
        else:
            return {key: value}
    except Exception as e:
        raise RuntimeError(f"Error formatting array property: {e}") from e


def construct_props(node_name: str, data: DataSourceProtocol) -> dict:
    """
    Construct the 'properties' section for a Gen3 schema node.

    This combines the node's own properties and any link properties.

    Args:
        node_name: The name of the node.
        data: The data structure containing nodes and links.

    Returns:
        A dictionary of all properties (including links) for the node.

    Raises:
        ValueError: If the node is not found.
    """
    links = get_node_links(node_name, data)
    props = get_properties(node_name, data)
    props = strip_required_field(props)
    node_data = get_node_data(node_name, data)
    category = node_data.category

    # Flatten property dicts into a single dict
    props_dict = {}
    # add ubiquitous property ref
    props_dict['$ref'] = "_definitions.yaml#/ubiquitous_properties"
    for prop in props:
        if isinstance(prop, dict):
            prop = format_enum(prop)
            prop = format_datetime(prop)
            prop = format_array(prop)
            props_dict.update(prop)

    # Add link properties
    for link in links:
        link_prop = create_link_prop(link['parent'], flip_multiplicity(link['multiplicity']))
        props_dict.update(link_prop)

    # if it's an Enum, add the enum values
    if category == "data_file":
        props_dict['core_metadata_collections'] = {"$ref": "_definitions.yaml#/to_one"}
        props_dict['$ref'] = "_definitions.yaml#/data_file_properties"

        props_dict['data_category'] = {
            "description": "Broad categorization of the contents of the data file.",
            "enum": ['data_category_1', 'data_category_2', 'data_category_3']
        }
        props_dict['data_format'] = {
            "description": "The format of the data in this data file",
            "enum": ['data_format_1', 'data_format_2', 'data_format_3']
        }
        props_dict['data_type'] = {
            "description": "The type of data in this data file",
            "enum": ['data_type_1', 'data_type_2', 'data_type_3']
        }

    return props_dict


def get_category(node_name: str, data: DataSourceProtocol) -> str:
    """
    Get the category value for a given node.

    Args:
        node_name: The name of the node.
        data: The data structure containing nodes.

    Returns:
        The category value (as a string).

    Raises:
        ValueError: If the node is not found.
    """
    ent = get_node_data(node_name, data)
    category = ent.category
    
    # If it's an Enum, get its value; otherwise, return as is
    if hasattr(category, "value"):
        return category.value
    return category


def get_node_value(node_name: str, key: str, data: DataSourceProtocol) -> str | list | dict | None:
    """
    Return the value of a single key within an node object.

    Args:
        node_name: The name of the node to retrieve.
        key: The key whose value is to be returned.
        data: The data structure containing nodes.

    Returns:
        The value associated with the specified key in the node object.

    Raises:
        ValueError: If the node is not found.
        KeyError: If the key doesn't exist in the node.
    """
    ent = get_node_data(node_name, data)
    node_dict = ent.model_dump()
    
    if key not in node_dict:
        raise KeyError(f"Key '{key}' not found in node '{node_name}'")
    
    return node_dict[key]


def is_file_node(node_name: str, data: DataSourceProtocol) -> bool:
    """
    Determine if an node is a file node (category: data_file).

    Args:
        node_name: The name of the node.
        data: The data structure containing nodes.

    Returns:
        True if the node is a file node, False otherwise.
    """
    try:
        category_value = get_node_value(node_name, 'category', data)
        return category_value == 'data_file'
    except (ValueError, KeyError):
        return False


def populate_template(node_name: str, input_data: DataSourceProtocol, template: dict) -> dict:
    """
    Populate a Gen3 schema template dictionary with values from a Pydantic data model.

    This function takes an node name, a Pydantic model instance containing node data,
    and a Gen3 schema template dictionary. It fills a copy of the template with values
    from the input data, applying special logic for certain keys (e.g., 'name', 'category',
    'properties', 'links').

    Args:
        node_name: The name of the node to populate in the template.
        input_data: A Pydantic model instance containing the node's data.
        template: A Gen3 schema template dictionary to be populated.

    Returns:
        A new Gen3 schema template dictionary populated with values from the input data.

    Raises:
        ValueError: If the node is not found in input_data.
    """
    ent = get_node_data(node_name, input_data)
    ent_dict = ent.model_dump()
    output_schema = template.copy()
    namespace = str(input_data.model_dump()['url'])
    
    # add node name as title
    output_schema['title'] = ent.name
    output_schema['namespace'] = namespace
    
    # Check if node is a file category
    file_node = is_file_node(node_name, input_data)
    
    # Populate template with node data
    for key, value in ent_dict.items():
        if key == 'name':
            output_schema['id'] = value
        elif key == 'category':
            output_schema[key] = get_category(node_name, input_data)
        elif key == 'properties':
            output_schema[key] = construct_props(node_name, input_data)
        elif key in output_schema:
            output_schema[key] = value
        else:
            logger.debug(f"Key '{key}' from node '{node_name}' not found in template")
    
    # add required props
    props = get_properties(node_name, input_data)
    required_props = get_required_prop_names(props)
    if required_props:
        required_props.append('submitter_id')
        required_props.append('type')
        output_schema['required'] = required_props
    
    # Process and add links
    links = get_node_links(node_name, input_data)
    converted_links = convert_node_links(links)
    
    # Add core metadata link for file nodes
    if file_node and links:
        converted_links = add_core_metadata_link(converted_links, links[0]['child'])
    
    # Create link group if multiple links exist
    if len(converted_links) > 1:
        output_schema['links'] = create_link_group(converted_links)
    else:
        output_schema['links'] = converted_links
    
    return output_schema



import logging
from gen3schemadev.schema.input_schema import DataModel
import yaml

logger = logging.getLogger(__name__)

def load_yaml(file_path):
    """
    Loads a YAML file and returns its contents as a Python object.

    Args:
        file_path (str): Path to the YAML file.

    Returns:
        dict: Parsed YAML content.

    Raises:
        FileNotFoundError: If the file does not exist.
        yaml.YAMLError: If the file is not valid YAML.
        Exception: For any other exceptions during file reading.
    """
    try:
        with open(file_path, 'r') as f:
            data = yaml.safe_load(f)
            logger.info(f"Successfully loaded YAML file: {file_path}")
            return data
    except FileNotFoundError as e:
        logger.error(f"File not found: {file_path}")
        raise
    except yaml.YAMLError as e:
        logger.error(f"YAML parsing error in file {file_path}: {e}")
        raise
    except Exception as e:
        logger.error(f"Unexpected error loading YAML file {file_path}: {e}")
        raise

def validate_input_yaml(file_path):
    """
    Validates the input YAML file against the Pydantic DataModel schema.

    Args:
        file_path (str): Path to the YAML file.

    Returns:
        DataModel: An instance of DataModel if validation is successful.

    Raises:
        ValidationError: If the data does not conform to the DataModel schema.
        Exception: For any other exceptions during validation.
    """
    try:
        data = load_yaml(file_path)
        validated = DataModel.model_validate(data)
        logger.info(f"Completed validation for file: {file_path}")
        return validated
    except Exception as e:
        logger.error(f"Failed to run validation for file {file_path}: {e}")
        raise



from typing import List, Literal, Optional
from enum import Enum 
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    AnyUrl,
)


class Property(BaseModel):
    """Schema for gen3 property"""
    name: str = Field(description="The name of the property.")
    type: Literal[
        'string', 'integer', 'number', 'boolean', 'datetime', 'enum', 'array'
    ] = Field(description="The data type of the property.")
    description: str = Field(description="A human-readable description of the property.")
    required: bool = Field(default=False, description="Whether this property is required for the node.")
    # Accepts either a list of EnumValue objects or a list of strings (for YAML input like in input_example.yml)
    enums: Optional[List[str]] = Field(
        default=None,
        description="A string list of possible values."
    )

    @field_validator('enums')
    def check_enums_for_enum_type(cls, v, info):
        """
        Validates that 'enums' is provided if 'type' is 'enum',
        matching the conditional requirement in the schema.
        """
        if info.data.get('type') == 'enum' and v is None:
            raise ValueError("The 'enums' field is required when the property type is 'enum'")
        return v

    class ConfigDict:
        extra = 'forbid'


# Inherit from str and Enum to create a string-based enumeration
class CategoryEnum(str, Enum):
    ADMINISTRATIVE = "administrative"
    ANALYSIS = "analysis"
    BIOSPECIMEN = "biospecimen"
    CLINICAL = "clinical"
    DATA_FILE = "data_file"
    METADATA_FILE = "metadata_file"
    NOTATION = "notation"
    INDEX_FILE = "index_file"
    CLINICAL_ASSESSMENT = "clinical_assessment"
    MEDICAL_HISTORY = "medical_history"
    DATA_OBSERVATIONS = "data_observations"
    EXPERIMENTAL_METHODS = "experimental_methods"
    SUBJECT_CHARACTERISTICS = "subject_characteristics"
    IMAGING = "imaging"
    STUDY_ADMINISTRATION = "study_administration"
    SATELLITE = "satellite"
    RADAR = "radar"
    STREAM_GAUGE = "stream_gauge"
    WEATHER_STATION = "weather_station"


class node(BaseModel):
    """A data node (node) in the model."""
    name: str = Field(description="The unique name of the node.")
    description: Optional[str] = Field(default=None, description="A human-readable description of the node.")
    category: CategoryEnum = Field(description="The category the node belongs to.")
    properties: List[Property] = Field(default_factory=list, description="A list of properties for the node.")
    
    class ConfigDict:
        extra = 'forbid'


class Link(BaseModel):
    """A relationship between nodes."""
    parent: str = Field(description="The parent node in the relationship.")
    multiplicity: Literal['one_to_many', 'many_to_one', 'one_to_one', 'many_to_many'] = Field(description="The cardinality of the relationship.")
    child: str = Field(description="The child node in the relationship.")
    
    class ConfigDict:
        extra = 'forbid'


class DataModel(BaseModel):
    """
    Schema to validate the data model configuration file.
    """
    version: str = Field(pattern=r'^\d+\.\d+\.\d+$', description="The semantic version of the configuration file.")
    url: AnyUrl = Field(description="The URL to the data portal.")
    nodes: List[node] = Field(description="A list of data nodes (nodes) in the model.")
    links: List[Link] = Field(description="A list of relationships between nodes.")

    class ConfigDict:
        extra = 'forbid'
